/*      xmms - jack output plugin
 *	Copyright 2002 Chris Morgan<cmorgan@alum.wpi.edu>
 *
 *	This code maps xmms calls into the jack translation library
 *
 *              Contains code Copyright (C) 1998-1999 Mikael Alm, Olle Hallnas,
 *              Thomas Nillson and 4Front Technologies
 */

#include "jackout.h"
#include "config.h"
#include "bio2jack.h" /* includes for the bio2jack library */
#include "xconvert.h" /* xmms rate conversion header file */

#define TRACE_ENABLE 0

#define OUTFILE stderr
#if TRACE_ENABLE
    #define TRACE(...) fprintf(OUTFILE, "%s:", __FUNCTION__),	\
          fprintf(OUTFILE, __VA_ARGS__),				\
                        fflush(OUTFILE);
    #define ERR(...) fprintf(OUTFILE, "ERR: %s:", __FUNCTION__),	\
          fprintf(OUTFILE, __VA_ARGS__),				\
                        fflush(OUTFILE);
#else
        #define TRACE(...) do{}while(0)
        #define ERR(...) do{}while(0)
#endif


static int driver = 0; /* handle to the jack output device */

jackconfig jack_cfg;

static int format, frequency, channels, bps;
static int input_format, input_frequency, input_channels, input_bps;
static int jack_server_rate = 0; /* true jack's rate */
static int xmms_rate = 0; /* rate */

static convert_freq_func_t freq_convert; /* rate convert function */
static struct xmms_convert_buffers *convertb; /* convert buffer */


/* set to 1 for verbose output */
#define VERBOSE_OUTPUT          0

#ifdef HAVE_CTOR
void destructor() __attribute__((destructor));
static gboolean output_opened; /* true if we have a connection to jack */
#endif


static GtkWidget *dialog, *button, *label;


#if HAVE_CTOR
void destructor()
{
  int errval;
  TRACE("cleanup\n");

  if((errval = JACK_Close(driver)))
  {
    ERR("error closing device, errval of %d\n", errval);
  }

  xmms_convert_buffers_destroy(convertb); /* clean up the rate conversion buffers */

  return;
}
#endif /* #if HAVE_CTOR */

#if 0
void jack_sample_rate_error(void)
{
	dialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dialog), ("Sample rate mismatch"));
	gtk_container_border_width(GTK_CONTAINER(dialog), 5);
	label = gtk_label_new((
		"Xmms is asking for a sample rate that differs from\n "
		"that of the jack server.  Until the xmms-jack plugin\n"
		"gets sample rate conversion the easiest solution is\n"
		"to change the sample rate that jack is running at.\n"
		"For the alsa driver '-r SAMPLE_RATE' will do.\n"
                "Patches to implement resampling are welcome ;-)\n\n"
		"Chris Morgan <cmorgan@alum.wpi.edu>\n"));

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	button = gtk_button_new_with_label((" Close "));
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
	gtk_widget_show(button);

	gtk_widget_show(dialog);
	gtk_widget_grab_focus(button);
}
#endif

/* Return the number of milliseconds of audio data that has been */
/* written out to the device */
gint jack_get_written_time(void)
{
  long return_val;
  return_val = JACK_GetPosition(driver, MILLISECONDS, WRITTEN);

  TRACE("returning %d milliseconds\n", return_val);
  return return_val;
}

/* Return the current number of milliseconds of audio data that has */
/* been played out of the audio device, not including the buffer */
gint jack_get_output_time(void)
{
  gint return_val;
  
  /* don't try to get any values if the device is still closed */
  if(JACK_GetState(driver) == CLOSED)
    return_val = 0;
  else
    return_val = JACK_GetPosition(driver, MILLISECONDS, PLAYED); /* get played position in terms of milliseconds */

  TRACE("returning %d milliseconds\n", return_val);
  return return_val;
}

/* returns TRUE if we are currently playing */
/* NOTE: this was confusing at first BUT, if the device is open and there */
/* is no more audio to be played, then the device is NOT PLAYING */
gint jack_playing(void)
{
  gint return_val;

  /* If we are playing see if we ACTUALLY have something to play */
  if(JACK_GetState(driver) == PLAYING)
  {
    /* If we have zero bytes stored, we are done playing */
    if(JACK_GetBytesStored(driver) == 0)
      return_val = FALSE;
    else
      return_val = TRUE;
  }
  else
    return_val = FALSE;

  TRACE("returning %d\n", return_val);
  return return_val;
}

/* Initialize necessary things */
void jack_init(void)
{
  TRACE("\n");
  JACK_Init(); /* initialize the driver */

  convertb = xmms_convert_buffers_new ();
  freq_convert = xmms_convert_get_frequency_func(FMT_S16_LE, 2);

#if HAVE_CTOR
  output_opened = FALSE;
#endif /* #if HAVE_CTOR */
}

/* Return the amount of data that can be written to the device */
gint jack_free(void)
{
  long return_val;
  return_val = JACK_GetBytesFreeSpace(driver);

  TRACE("free space of %d bytes\n", return_val);
  return return_val;
}

/* write some audio out to the device */
void jack_write(gpointer ptr, gint length)
{
  long written;
  EffectPlugin *ep;
  AFormat new_format;
  int new_frequency, new_channels;
  long positionMS;

  /* copy the current values into temporary values */
  new_format = input_format;
  new_frequency = input_frequency;
  new_channels = input_channels;

  /* query xmms for the current plugin */
  ep = get_current_effect_plugin();
  if(effects_enabled() && ep && ep->query_format)
  {
    ep->query_format(&new_format, &new_frequency, &new_channels);
  }

  /* if the format has changed take this into account by modifying */
  /* the time offset and reopening the device with the new format settings */
  if (new_format != (AFormat)format ||
      new_frequency != frequency ||
      new_channels != channels)
  {
    TRACE("format changed, storing new values and opening/closing jack\n");
 
    positionMS = JACK_GetPosition(driver, MILLISECONDS, PLAYED);

    /* store the new values */
    format = new_format;
    frequency = new_frequency;
    channels = new_channels;
    bps = channels * frequency;

    jack_close();
    jack_open(new_format, new_frequency, new_channels);
 
    /* restore the position after the open and close */
    JACK_SetState(driver, PAUSED);
    JACK_SetPosition(driver, MILLISECONDS, positionMS);
    JACK_SetState(driver, PLAYING);
  }

  /* if we need rate conversion, perform it here */
  if (jack_server_rate != xmms_rate)
  {
     length = freq_convert (convertb, &ptr, length, xmms_rate, jack_server_rate);
  }

  /* if effects are enabled and we have a plugin, run the current */
  /* samples through the plugin */
  if (effects_enabled() && ep && ep->mod_samples)
  {
           length = ep->mod_samples(&ptr, length,
                                 input_format,
                                  input_frequency,
                                  input_channels);
  }

  TRACE("legnth = %d\n", length);
  /* loop until we have written all the data out to the jack device */
  /* this is due to xmms' audio driver api */
  while(length > 0)
  {
    TRACE("writing %d bytes\n", length);
    written = JACK_Write(driver, ptr, length);
    length-=written;
    (char *)ptr += written;
  }
  TRACE("finished\n");
}

/* Close the device */
void jack_close(void)
{
  int errval;

  TRACE("\n");

#if HAVE_CTOR
  JACK_Reset(driver); /* flush buffers, reset position and set state to STOPPED */
  TRACE("resetting driver, not closing now, destructor will close for us\n");
#else
  if((errval = JACK_Close(driver)))
  {
    ERR("error closing device, errval of %d\n", errval);
  }
#endif /* #if HAVE_CTOR */
}

/* Flush any output currently buffered */
/* and set the number of bytes written based on ms_offset_time, */
/* the number of milliseconds of offset passed in */
/* This is done so the driver itself keeps track of */
/* current playing position of the mp3 */
void jack_flush(gint ms_offset_time)
{
  TRACE("setting values for ms_offset_time of %d\n", ms_offset_time);

  JACK_Reset(driver); /* flush buffers and set state to STOPPED */

  /* update the internal driver values to correspond to the input time given */
  JACK_SetPosition(driver, MILLISECONDS, ms_offset_time);

  JACK_SetState(driver, PLAYING);
}

/* Pause the jack device */
void jack_pause(short p)
{
  TRACE("p == %d\n", p);

  /* pause the device if p is non-zero, unpause the device if p is zero and */
  /* we are currently paused */
  if(p)
    JACK_SetState(driver, PAUSED);
  else if(JACK_GetState(driver) == PAUSED)
    JACK_SetState(driver, PLAYING);
}

/* Open the device up */
gint jack_open(AFormat fmt, gint sample_rate, gint num_channels)
{
  int bits_per_sample;
  int retval;

  TRACE("\n");

  xmms_rate = sample_rate; /* what rate does xmms want? */

  if((fmt == FMT_U8) || (fmt == FMT_S8))
  {
    bits_per_sample = 8;
  } else
  {
    bits_per_sample = 16;
  }

  /* record some useful information */
  input_format = format;
  input_channels = channels;
  input_frequency = frequency;
  input_bps = bps;

#if HAVE_CTOR
  /* if we are already opened then don't open again */
  if(output_opened)
  {
    TRACE("output_opened is TRUE, not reopening\n");
    return 1;
  }
#endif /* #if HAVE_CTOR */

  /* try to open the jack device with the xmms_rate at first */
  jack_server_rate = xmms_rate;
  retval = JACK_Open(&driver, bits_per_sample, &jack_server_rate, num_channels);
  if(retval == ERR_RATE_MISMATCH)
  {
    TRACE("xmms wants rate of '%d', jacks rate is '%d', opening at jack rate\n", xmms_rate, jack_server_rate);
    /* open the jack device with true jack's rate, return 0 upon failure */
    if(retval = JACK_Open(&driver, bits_per_sample, &jack_server_rate, num_channels))
    {
      TRACE("failed to open jack with JACK_Open(), error %d\n", retval);
      return 0;
    }
    TRACE("success!!\n");
  } else if(retval != ERR_SUCCESS)
  {
    TRACE("failed to open jack with JACK_Open(), error %d\n", retval);
    return 0;
  }

#if HAVE_CTOR
  output_opened = TRUE;
#endif

  return 1;
}

/* Set the volume */
void jack_set_volume(int l, int r)
{
  TRACE("l(%d), r(%d)\n", l, r);
  JACK_SetVolume(driver, l, r);
}

/* Get the current volume setting */
void jack_get_volume(int *l, int *r)
{
  JACK_GetVolume(driver, l, r);

#if VERBOSE_OUTPUT
  TRACE("l(%d), r(%d)\n", *l, *r);
#endif
}



